@@ -98,6 +98,7 @@ public class Driver {
   private UserGroupInformation ugi;
   private FileSystem defaultFileSystem;
   private YarnClient defaultYarnClient;
+  private LogClient logClient;
 
   private String classpath;
 
@@ -156,7 +157,7 @@ public class Driver {
   private void stopServer() {
     if (server != null) {
       server.shutdown();
-      LOG.info("Driver shut down");
+      LOG.debug("Driver shut down");
     }
   }
 
@@ -279,6 +280,8 @@ public class Driver {
     defaultFileSystem = getFs();
     // Start the yarn client as *this* user
     defaultYarnClient = getYarnClient();
+    // Create a logs client
+    logClient = new LogClient(conf);
 
     // Start the server
     startServer();
@@ -315,6 +318,21 @@ public class Driver {
     return new Path(fs.getHomeDirectory(), ".skein/" + appId.toString());
   }
 
+  public Map<String, String> getApplicationLogs(
+      final ApplicationId appId, final String owner, String user)
+      throws IOException, InterruptedException {
+    if (user.isEmpty()) {
+      return logClient.getLogs(appId, owner);
+    } else {
+      return UserGroupInformation.createProxyUser(user, ugi).doAs(
+          new PrivilegedExceptionAction<Map<String, String>>() {
+            public Map<String, String> run() throws IOException {
+              return logClient.getLogs(appId, owner);
+            }
+          });
+    }
+  }
+
   public void killApplication(final ApplicationId appId, String user)
       throws IOException, YarnException, InterruptedException {
     if (user.isEmpty()) {
@@ -676,6 +694,17 @@ public class Driver {
     }
   }
 
+  private boolean hasCompleted(ApplicationReport report) {
+    switch (report.getYarnApplicationState()) {
+      case FINISHED:
+      case FAILED:
+      case KILLED:
+        return true;
+      default:
+        return false;
+    }
+  }
+
   private synchronized void addWatcher(final ApplicationId appId,
       StreamObserver<Msg.ApplicationReport> resp) {
 
@@ -1040,6 +1069,49 @@ public class Driver {
       }
     }
 
+    @Override
+    public void getLogs(Msg.LogsRequest req,
+        StreamObserver<Msg.LogsResponse> resp) {
+
+      if (notLoggedIn(resp)) {
+        return;
+      }
+
+      ApplicationReport report = getReport(req.getId(), resp);
+      if (report == null) {
+        return;
+      }
+
+      Map<String, String> logs;
+      if (hasCompleted(report)) {
+        try {
+          logs = getApplicationLogs(
+              report.getApplicationId(), report.getUser(), req.getUser());
+        } catch (LogClient.LogClientException exc) {
+          resp.onError(Status.INVALID_ARGUMENT
+              .withDescription(exc.getMessage())
+              .asRuntimeException());
+          return;
+        } catch (Exception exc) {
+          resp.onError(Status.INTERNAL
+              .withDescription("Failed to get logs for application '"
+                               + req.getId()
+                               + "', exception:\n"
+                               + exc.getMessage())
+              .asRuntimeException());
+          return;
+        }
+      } else {
+        resp.onError(Status.INVALID_ARGUMENT
+            .withDescription("Application " + req.getId()
+                             + " has not completed, logs are not available")
+            .asRuntimeException());
+        return;
+      }
+      resp.onNext(Msg.LogsResponse.newBuilder().putAllLogs(logs).build());
+      resp.onCompleted();
+    }
+
     @Override
     public void waitForStart(Msg.Application req,
         StreamObserver<Msg.ApplicationReport> resp) {
@@ -1069,7 +1141,15 @@ public class Driver {
         return;
       }
 
-      Model.ApplicationSpec spec = MsgUtils.readApplicationSpec(req);
+      Model.ApplicationSpec spec;
+      try {
+        spec = MsgUtils.readApplicationSpec(req);
+      } catch (Utils.UnsupportedFeatureException exc) {
+        resp.onError(Status.INVALID_ARGUMENT
+            .withDescription(exc.getMessage())
+            .asRuntimeException());
+        return;
+      }
 
       ApplicationId appId = null;
       try {
