@@ -1,5 +1,3 @@
-from __future__ import print_function, division, absolute_import
-
 import datetime
 import errno
 import json
@@ -11,11 +9,11 @@ import struct
 import subprocess
 import warnings
 from contextlib import closing
+from collections.abc import Mapping
 
 import grpc
 
 from . import proto
-from .compatibility import PY2, makedirs, isidentifier, Mapping
 
 from .exceptions import (context, ConnectionError, TimeoutError,
                          ApplicationNotRunningError, ApplicationError,
@@ -25,7 +23,7 @@ from .ui import WebUI
 from .model import (Security, ApplicationSpec, ApplicationReport,
                     ApplicationState, ContainerState, Container,
                     FinalStatus, Resources, container_instance_from_string,
-                    LogLevel, NodeState, NodeReport, Queue)
+                    LogLevel, NodeState, NodeReport, Queue, ApplicationLogs)
 from .utils import (cached_property, grpc_fork_support_disabled, pid_exists,
                     datetime_to_millis)
 
@@ -118,7 +116,7 @@ class Properties(Mapping):
     def __dir__(self):
         o = set(dir(type(self)))
         o.update(self.__dict__)
-        o.update(c for c in self._mapping if isidentifier(c))
+        o.update(c for c in self._mapping if c.isidentifier())
         return list(o)
 
     def __iter__(self):
@@ -154,7 +152,7 @@ def _read_driver():
 
 def _write_driver(address, pid):
     # Ensure the config dir exists
-    makedirs(properties.config_dir, exist_ok=True)
+    os.makedirs(properties.config_dir, exist_ok=True)
     # Write to the driver file
     with open(os.path.join(properties.config_dir, 'driver'), 'w') as fil:
         json.dump({'address': address, 'pid': pid}, fil)
@@ -228,18 +226,10 @@ def _start_driver(security=None, set_global=False, keytab=None, principal=None,
         _, callback_port = callback.getsockname()
         env['SKEIN_CALLBACK_PORT'] = str(callback_port)
 
-        if PY2:
-            popen_kwargs = dict(preexec_fn=os.setsid)
-        else:
-            popen_kwargs = dict(start_new_session=True)
-
         if log is None:
             outfil = None
         elif log is False:
-            if PY2:
-                outfil = open(os.devnull, 'w')
-            else:
-                outfil = subprocess.DEVNULL
+            outfil = subprocess.DEVNULL
         else:
             outfil = open(log, mode='w')
         infil = None if set_global else subprocess.PIPE
@@ -249,7 +239,7 @@ def _start_driver(security=None, set_global=False, keytab=None, principal=None,
                                 stdout=outfil,
                                 stderr=outfil,
                                 env=env,
-                                **popen_kwargs)
+                                start_new_session=True)
 
         while proc.poll() is None:
             readable, _, _ = select.select([callback], [], [], 1)
@@ -335,7 +325,7 @@ class Client(_ClientBase):
     >>> with skein.Client() as client:
     ...     app_id = client.submit('spec.yaml')
     """
-    __slots__ = ('address', 'security', '_stub', '_proc')
+    __slots__ = ('address', 'security', '_channel', '_stub', '_proc')
     _server_name = 'driver'
     _server_error = DriverError
 
@@ -355,7 +345,8 @@ class Client(_ClientBase):
             proc = None
 
         with grpc_fork_support_disabled():
-            self._stub = proto.DriverStub(secure_channel(address, security))
+            self._channel = secure_channel(address, security)
+            self._stub = proto.DriverStub(self._channel)
         self.address = address
         self.security = security
         self._proc = proc
@@ -369,6 +360,9 @@ class Client(_ClientBase):
                 proc.wait()
             raise
 
+    def __reduce__(self):
+        return (type(self), (self.address, self.security))
+
     @classmethod
     def from_global_driver(self):
         """Connect to the global driver."""
@@ -477,10 +471,15 @@ class Client(_ClientBase):
         return 'Client<%s>' % self.address
 
     def close(self):
-        """Closes the java driver if started by this client. No-op otherwise."""
+        """Release all resources used by this client. Idempotent.
+
+        Note that this closes the Java driver if it was started by this client.
+        """
         if self._proc is not None:
             self._proc.stdin.close()
             self._proc.wait()
+        if hasattr(self, '_channel'):
+            self._channel.close()
 
     def __enter__(self):
         return self
@@ -795,6 +794,31 @@ class Client(_ClientBase):
         resp = self._call('getStatus', proto.Application(id=app_id))
         return ApplicationReport.from_protobuf(resp)
 
+    def application_logs(self, app_id, user=""):
+        """Get logs from a completed skein application.
+
+        Parameters
+        ----------
+        app_id : str
+            The id of the application.
+        user : str, optional
+            The user to get the application logs as. Requires the current user
+            to have permissions to proxy as ``user``. Default is the current
+            user.
+
+        Returns
+        -------
+        ApplicationLogs : logs
+            A mapping of ``yarn_container_id`` to ``logs`` for each container.
+
+        Examples
+        --------
+        >>> client.application_logs('application_1526134340424_0012')
+        ApplicationLogs<application_1526134340424_0012>
+        """
+        resp = self._call('getLogs', proto.LogsRequest(id=app_id, user=user))
+        return ApplicationLogs(app_id, dict(resp.logs))
+
     def move_application(self, app_id, queue):
         """Move an application to a different queue.
 
@@ -844,11 +868,30 @@ class ApplicationClient(_ClientBase):
         self.security = security or Security.from_default()
         self.id = app_id
         with grpc_fork_support_disabled():
-            self._stub = proto.AppMasterStub(secure_channel(address, self.security))
+            self._channel = secure_channel(address, self.security)
+            self._stub = proto.AppMasterStub(self._channel)
 
     def __repr__(self):
         return 'ApplicationClient<%s, %s>' % (self.id, self.address)
 
+    def close(self):
+        """Release all resources used by this client. Idempotent.
+
+        Note that this only frees resources used by the client, it doesn't
+        shutdown the application itself. For that see ``shutdown``.
+        """
+        if hasattr(self, '_channel'):
+            self._channel.close()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args):
+        self.close()
+
+    def __del__(self):
+        self.close()
+
     def shutdown(self, status='SUCCEEDED', diagnostics=None):
         """Shutdown the application.
 
@@ -962,6 +1005,30 @@ class ApplicationClient(_ClientBase):
         resp = self._call('scale', req)
         return [Container.from_protobuf(c) for c in resp.containers]
 
+    def add_container(self, service, env=None):
+        """Add a new container to a service.
+
+        Unlike ``scale``, this adds the ability to override configuration of
+        the requested container.
+
+        Parameters
+        ----------
+        service : str
+            The service to scale.
+        env : dict, optional
+            A mapping of environment variables to set in the container. These
+            will be applied after any environment variables set in the service
+            description, and can be used as overrides.
+
+        Returns
+        -------
+        container : Container
+            The new container that was started.
+        """
+        req = proto.AddContainerRequest(service_name=service, env=env)
+        resp = self._call('addContainer', req)
+        return Container.from_protobuf(resp)
+
     def set_progress(self, progress):
         """Update the progress for this application.
 
