@@ -4,10 +4,11 @@ from .skein_pb2 import (Empty, FinalStatus, ApplicationState, Resources, File,
                         Service, Acls, Log, Master, Security, ApplicationSpec,
                         ResourceUsageReport, ApplicationReport, Application,
                         ApplicationsRequest, Url, ContainersRequest, Container,
-                        ContainerInstance, ScaleRequest, ShutdownRequest,
-                        KillRequest, SetProgressRequest, NodeState, NodeReport,
-                        NodesRequest, Queue, QueueRequest, QueuesResponse,
-                        MoveRequest)
+                        ContainerInstance, ScaleRequest, AddContainerRequest,
+                        ShutdownRequest, KillRequest, SetProgressRequest,
+                        NodeState, NodeReport, NodesRequest, Queue,
+                        QueueRequest, QueuesResponse, MoveRequest, LogsRequest,
+                        LogsResponse)
 from .skein_pb2 import (GetRangeRequest, GetRangeResponse,
                         PutKeyRequest, PutKeyResponse,
                         DeleteRangeRequest, DeleteRangeResponse,
