@@ -1,7 +1,6 @@
-from __future__ import print_function, division, absolute_import
-
 import datetime
 import os
+import pickle
 import subprocess
 import time
 import warnings
@@ -16,7 +15,8 @@ from skein.core import Properties
 from skein.exceptions import FileNotFoundError, FileExistsError
 from skein.utils import pid_exists
 from skein.test.conftest import (run_application, wait_for_containers,
-                                 wait_for_completion, get_logs, KEYTAB_PATH)
+                                 wait_for_completion, get_logs, KEYTAB_PATH,
+                                 HADOOP3)
 
 
 def test_properties():
@@ -196,6 +196,7 @@ def test_client_errors_nicely_if_not_logged_in(security, not_logged_in):
                            ('get_child_queues', ('default',)),
                            ('get_all_queues', ()),
                            ('application_report', (appid,)),
+                           ('application_logs', (appid,)),
                            ('connect', (appid,)),
                            ('move_application', (appid, 'default')),
                            ('kill_application', (appid,)),
@@ -276,6 +277,16 @@ def test_client_login_from_keytab(security, not_logged_in):
         skein.Client(keytab=KEYTAB_PATH, security=security)
 
 
+def test_client_picklable(client):
+    b = pickle.dumps(client)
+    client2 = pickle.loads(b)
+    assert client2.address == client.address
+    assert client2.security == client.security
+    assert client2._proc is None
+    # Smoketest
+    client2.get_applications()
+
+
 def test_get_nodes(client):
     nodes = client.get_nodes()
     assert nodes
@@ -326,16 +337,21 @@ def test_get_applications(client, at_least_3_apps_in_history):
     all_expected = [a for a in all_apps if a.user == 'testuser' and a.queue == 'default']
     assert all_expected
 
+    def assert_eq(a, b):
+        a = set(i.id for i in a)
+        b = set(i.id for i in b)
+        assert a == b
+
     # Filter on user and queue to show identical
     common = {'states': all_states, 'user': 'testuser', 'queue': 'default'}
     res = client.get_applications(**common)
-    assert res == all_expected
+    assert_eq(res, all_expected)
 
     # Filter on name as well
     app_name = all_expected[0].name
     sol = [a for a in all_expected if a.name == app_name]
     res = client.get_applications(name=app_name, **common)
-    assert res == sol
+    assert_eq(res, sol)
 
     # Filter out all apps
     assert not client.get_applications(states=all_states, user='not-a-real-value')
@@ -349,22 +365,22 @@ def test_get_applications(client, at_least_3_apps_in_history):
     # All apps started after start_time
     sol = [a for a in all_expected if a.start_time >= start_time]
     res = client.get_applications(started_begin=start_time, **common)
-    assert res == sol
+    assert_eq(res, sol)
 
     # All apps started before start_time
     sol = [a for a in all_expected if a.start_time <= start_time]
     res = client.get_applications(started_end=start_time, **common)
-    assert res == sol
+    assert_eq(res, sol)
 
     # All apps finished after finish_time
     sol = [a for a in all_expected if a.finish_time >= finish_time]
     res = client.get_applications(finished_begin=finish_time, **common)
-    assert res == sol
+    assert_eq(res, sol)
 
     # All apps finished before finish_time
     sol = [a for a in all_expected if a.finish_time <= finish_time]
     res = client.get_applications(finished_end=finish_time, **common)
-    assert res == sol
+    assert_eq(res, sol)
 
     # Check time parsing
     t = datetime.datetime.now()
@@ -386,7 +402,7 @@ def test_get_applications(client, at_least_3_apps_in_history):
     res = client.get_applications(
         finished_end=finish_time.strftime('%Y-%m-%d %H:%M:%S'), **common
     )
-    assert res == sol
+    assert_eq(res, sol)
 
     with pytest.raises(TypeError) as exc:
         res = client.get_applications(finished_end=123)
@@ -464,6 +480,17 @@ def test_simple_app(client):
 
         assert client.application_report(app.id).state == 'RUNNING'
 
+        # Connect with a new client
+        with client.connect(app.id) as app2:
+            app2.get_specification()
+
+        # application client is closed
+        with pytest.raises(Exception):
+            app2.get_specification()
+
+        # Closing an application client is idempotent
+        app2.close()
+
         app.shutdown()
 
     with pytest.raises(skein.ConnectionError):
@@ -626,6 +653,61 @@ def test_dynamic_containers(client):
         app.shutdown()
 
 
+def test_add_container(client):
+    script = ('echo "$SKEIN_CONTAINER_ID - MYENV=$MYENV"\n'
+              'echo "$SKEIN_CONTAINER_ID - MYENV2=$MYENV2"\n'
+              'if [[ "$MYENV" == "bar" ]]; then\n'
+              '  exit 1\n'
+              'else\n'
+              '  exit 0\n'
+              'fi')
+
+    spec = skein.ApplicationSpec(
+        name="test_add_container",
+        master=skein.Master(script="sleep infinity"),
+        services={
+            'test': skein.Service(
+                instances=0,
+                resources=skein.Resources(memory=32, vcores=1),
+                env={'MYENV': 'foo',
+                     'MYENV2': 'baz'},
+                max_restarts=1,
+                script=script
+            )
+        }
+    )
+
+    with run_application(client, spec=spec) as app:
+        # Add container with new overrides
+        c = app.add_container('test')
+        assert c.instance == 0
+        wait_for_containers(app, 1, states=['RUNNING', 'SUCCEEDED'])
+
+        # Non-existant service
+        with pytest.raises(ValueError):
+            app.add_container('foobar')
+
+        # Add container with override for MYENV
+        c = app.add_container('test', {'MYENV': 'bar'})
+        assert c.instance == 1
+
+        # The new env var triggers a failure, should fail twice,
+        # then fail the whole application
+        assert wait_for_completion(client, app.id) == 'FAILED'
+
+    logs = get_logs(client, app.id)
+    assert "test_0 - MYENV=foo" in logs
+    assert "test_0 - MYENV2=baz" in logs
+
+    assert "test_1 - MYENV=bar" in logs
+    assert "test_1 - MYENV2=baz" in logs
+
+    assert "test_2 - MYENV=bar" in logs
+    assert "test_2 - MYENV2=baz" in logs
+
+    assert "test_3" not in logs
+
+
 @pytest.mark.parametrize('runon', ['service', 'master'])
 def test_container_environment(runon, client, has_kerberos_enabled):
     script = ('set -e\n'
@@ -649,7 +731,7 @@ def test_container_environment(runon, client, has_kerberos_enabled):
     with run_application(client, spec=spec, connect=False) as app_id:
         assert wait_for_completion(client, app_id) == 'SUCCEEDED'
 
-    logs = get_logs(app_id)
+    logs = get_logs(client, app_id)
     assert "USER=testuser" in logs
     assert 'SKEIN_APPMASTER_ADDRESS=' in logs
     assert 'SKEIN_APPLICATION_ID=%s' % app_id in logs
@@ -745,7 +827,7 @@ def test_custom_log4j_properties(client, tmpdir):
     with run_application(client, spec=spec) as app:
         assert wait_for_completion(client, app.id) == 'SUCCEEDED'
 
-    logs = get_logs(app.id)
+    logs = get_logs(client, app.id)
     assert 'CUSTOM-LOG4J-SUCCEEDED' in logs
 
 
@@ -760,11 +842,16 @@ def test_set_log_level(client):
     with run_application(client, spec=spec) as app:
         assert wait_for_completion(client, app.id) == 'SUCCEEDED'
 
-    logs = get_logs(app.id)
+    logs = get_logs(client, app.id)
     assert 'DEBUG' in logs
 
 
-@pytest.mark.parametrize('kind', ['master', 'service'])
+@pytest.mark.parametrize('kind', [
+    pytest.param('master', marks=pytest.mark.xfail(
+        reason="Memory error on master isn't deterministic"
+    )),
+    'service'
+])
 def test_memory_limit_exceeded(kind, client):
     resources = skein.Resources(memory=32, vcores=1)
     # Allocate noticeably more memory than the 32 MB limit
@@ -785,7 +872,7 @@ def test_memory_limit_exceeded(kind, client):
                                  services=services)
     with run_application(client, spec=spec, connect=False) as app_id:
         assert wait_for_completion(client, app_id) == "FAILED"
-    logs = get_logs(app_id)
+    logs = get_logs(client, app_id)
     assert search_txt in logs
 
     if kind == 'master':
@@ -856,7 +943,7 @@ def test_proxy_user(client):
         services={
             "service": skein.Service(
                 resources=skein.Resources(memory=32, vcores=1),
-                script='sleep infinity')
+                script="sleep infinity")
         }
     )
     with run_application(client, spec=spec) as app:
@@ -868,6 +955,11 @@ def test_proxy_user(client):
     for fil in spec2.services['service'].files.values():
         assert fil.source.startswith('hdfs://master.example.com:9000/user/alice')
 
+    # Can get logs as user
+    logs = get_logs(client, app.id, user="alice")
+    assert app.id in logs
+    assert "application.master.log" in logs
+
     # Application directory deleted after kill
     fs = hdfs.connect()
     assert not fs.exists("/user/testuser/.skein/%s" % app.id)
@@ -942,7 +1034,7 @@ def test_master_driver_foo(client, tmpdir):
     with run_application(client, spec=spec, connect=False) as app_id:
         assert wait_for_completion(client, app_id) == 'SUCCEEDED'
 
-    logs = get_logs(app_id)
+    logs = get_logs(client, app_id)
     assert 'FOO=BAR' in logs
     assert 'myfile' in logs
 
@@ -1006,7 +1098,7 @@ def test_retries_succeeds(client):
     )
     with run_application(client, spec=spec, connect=False) as app_id:
         assert wait_for_completion(client, app_id) == 'SUCCEEDED'
-    logs = get_logs(app_id)
+    logs = get_logs(client, app_id)
     assert 'Failing on other attempts' in logs
     assert 'Application attempt 1 out of 2 failed, will retry' in logs
     assert 'Succeeding on attempt 02' in logs
@@ -1030,7 +1122,7 @@ def test_retries_fails(client):
     )
     with run_application(client, spec=spec, connect=False) as app_id:
         assert wait_for_completion(client, app_id) == 'FAILED'
-    logs = get_logs(app_id)
+    logs = get_logs(client, app_id)
     assert logs.count('Failing on other attempts') == 2
     assert 'Application attempt 1 out of 2 failed' in logs
 
@@ -1071,6 +1163,36 @@ def test_allow_failures_max_restarts(client, allow_failures):
             assert wait_for_completion(client, app.id) == 'FAILED'
 
 
+@pytest.mark.parametrize('with_restarts', [True, False])
+def test_fail_on_container_failure(client, with_restarts):
+    script = ('if [[ "$SKEIN_CONTAINER_ID" != "test_0" ]]; then\n'
+              '  exit 1\n'
+              'else\n'
+              '  sleep infinity\n'
+              'fi')
+
+    spec = skein.ApplicationSpec(
+        name="test_fail_on_container_failure",
+        services={
+            'test': skein.Service(
+                instances=2,
+                max_restarts=2 if with_restarts else 0,
+                resources=skein.Resources(memory=32, vcores=1),
+                script=script
+            )
+        }
+    )
+    with run_application(client, spec=spec) as app:
+        wait_for_completion(client, app.id) == "FAILED"
+
+    logs = get_logs(client, app.id)
+    assert "test_0" in logs
+    assert "test_1" in logs
+    assert ("test_2" in logs) == with_restarts
+    assert ("test_3" in logs) == with_restarts
+    assert "test_4" not in logs
+
+
 def test_move_application(client):
     spec = skein.ApplicationSpec(
         name="test_move_application",
@@ -1121,3 +1243,23 @@ def test_move_application(client):
     with pytest.raises(ValueError) as exc:
         client.move_application("oh no", "default")
     assert "Invalid" in str(exc.value)
+
+
+def test_hadoop3_resource(client):
+    spec = skein.ApplicationSpec(
+        name="test_hadoop3_resources",
+        master=skein.Master(
+            resources=skein.Resources(
+                memory='32 MiB',
+                vcores=1,
+                gpus=1
+            ),
+            script="sleep infinity"
+        )
+    )
+    with pytest.raises(ValueError) as exc:
+        client.submit(spec)
+    if HADOOP3:
+        assert "Resource 'yarn.io/gpu'" in str(exc.value)
+    else:
+        assert "Custom resources not supported"
