@@ -1,17 +1,16 @@
-from __future__ import print_function, division, absolute_import
-
 import copy
 import datetime
+import math
 import os
 import pickle
 
 import pytest
 
-from skein.compatibility import math_ceil
 from skein.model import (ApplicationSpec, Service, Resources, File,
                          ApplicationState, FinalStatus, FileType, ACLs, Master,
                          Container, ApplicationReport, ResourceUsageReport,
-                         NodeReport, LogLevel, parse_memory, Security, Queue)
+                         NodeReport, LogLevel, parse_memory, Security, Queue,
+                         ApplicationLogs)
 
 
 def indent(s, n):
@@ -95,14 +94,14 @@ def test_parse_memory():
     mib = 2 ** 20
     assert parse_memory('100') == 100
     assert parse_memory('100 MiB') == 100
-    assert parse_memory('100 MB') == math_ceil(100 * 1e6 / mib)
-    assert parse_memory('100M') == math_ceil(100 * 1e6 / mib)
+    assert parse_memory('100 MB') == math.ceil(100 * 1e6 / mib)
+    assert parse_memory('100M') == math.ceil(100 * 1e6 / mib)
     assert parse_memory('100Mi') == 100
     assert parse_memory('5kB') == 1
     assert parse_memory('5.4 MiB') == 6
     assert parse_memory('0.9 MiB') == 1
     assert parse_memory('1e3') == 1000
-    assert parse_memory('1e6 kB') == math_ceil(1e6 * 1e3 / mib)
+    assert parse_memory('1e6 kB') == math.ceil(1e6 * 1e3 / mib)
     assert parse_memory('MiB') == 1
 
     with pytest.raises(ValueError):
@@ -127,7 +126,9 @@ def test_parse_memory():
 def test_resources():
     r = Resources(memory=1024, vcores=1)
     r2 = Resources(memory=1024, vcores=2)
+    r3 = Resources(memory=1, vcores=2, gpus=3, fpgas=4)
     check_specification_methods(r, r2)
+    check_specification_methods(r3, r2)
 
 
 def test_resources_invariants():
@@ -685,3 +686,20 @@ def test_queue():
     a = Queue('queue1', 'RUNNING', 50.0, 60.0, 80.0, set(), '')
     b = Queue('queue2', 'RUNNING', 40.0, 60.0, 30.0, set('*'), '')
     check_base_methods(a, b)
+
+
+def test_application_logs():
+    raw = {"container_1528138529205_0038_01_000001": "line 1\nline 2\nline 3",
+           "container_1528138529205_0038_01_000002": "line 4\nline 5\nline 6"}
+
+    logs = ApplicationLogs('application_1528138529205_0001', raw)
+    assert repr(logs) == "ApplicationLogs<application_1528138529205_0001>"
+    assert len(logs) == 2
+    assert dict(logs) == raw
+    out = logs.dumps()
+    assert "line 6" in out
+    assert logs.app_id in out
+    assert all(k in out for k in raw)
+
+    html = logs._repr_html_()
+    assert logs.app_id in html
